/*                                                                                                            
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 */

/*
 *   Copyright (c) 2003 EU DataGrid        http://www.eu-datagrid.org/
 *
 *   $Id: pdl_lex.l 14868 2011-03-04 20:55:13Z dennisvd $
 *
 *   Copyright (c) 2003 by
 *      G.M. Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *   This software is distributed under a BSD-style open source
 *   licence. For a complete description of the licence take a look
 *   at: http://eu-datagrid.web.cern.ch/eu-datagrid/license.html
 *
 */


%{
#include <string.h>

#include "pdl.h"
#include "pdl_rule.h"
#include "pdl_yacc.h"
#include "pdl_policy.h"

/*static int no_valid_rule=0;*/

%}

ws           [\t ]+
nl           \n
term         [a-zA-Z_.0-9]+
label        {term}:
trans        "->"
comment      #
tokens       [\~=\|]
pvar         "path"
%{
/* subpath matches 0 or more . followed by / and non-whitespace, this matches
 * ./aap as ./ 
 * Under 'term' it would only match . so path gets precedence. Note further that
 * the original [\/\.][^\/ \t\n]* is not the same as that would only match the
 * single . at the beginning, like term, in which case 'term' gets precendence */
%}
path         \.*\/[^\/ \t\n]*
pdlstr       \"[^\"\n]*[\"\n]


%x linecom path pathcom mq
%%

{ws}         ;
{pvar}       { return PVAR; }
{term}       { if ((yylval.record = (record_t*)malloc(sizeof(record_t))))
                 if (((yylval.record)->string = strdup(yytext))) {
                   (yylval.record)->lineno = lineno;
                   return TERM;
                 }
               lcmaps_warning(PDL_ERROR, "out of memory.");
               return 0;
             }
{label}      { yytext[yyleng-1] = '\0';   /* remove trailing ':' */

               /*  Check if this policy rule is allowed or not. */
               if (lcmaps_allowed_policy_rule(yytext)) {
                 lcmaps_allow_rules(TRUE);
               } else {
                 /*  This label should not be added, however normal parsing
                  *  must continue.  */
                 lcmaps_allow_rules(FALSE);
               }
               if ((yylval.record = (record_t*)malloc(sizeof(record_t))))
                 if (((yylval.record)->string = (char *)strdup(yytext))) {
                   (yylval.record)->lineno = lineno;
                   return LABEL;
                 }
               lcmaps_warning(PDL_ERROR, "out of memory.");
               return 0; 
             }
{path}       { if ((yylval.record = (record_t *)malloc(sizeof(record_t))))
                 if (((yylval.record)->string = strdup(yytext))) {
                   (yylval.record)->lineno = lineno;
                   BEGIN path;
                   return PATH;
                 }
               lcmaps_warning(PDL_ERROR, "out of memory.");
               return 0; 
             }
<path>{path} { if ((yylval.record = (record_t *)malloc(sizeof(record_t))))
                 if (((yylval.record)->string = strdup(yytext))) {
                   (yylval.record)->lineno = lineno;
                   return PATH;
                 }
               lcmaps_warning(PDL_ERROR, "out of memory.");
               return 0; 
             }
<path>#      { BEGIN pathcom; }
<path>[ \t]* ;
<path>\n     { BEGIN INITIAL; yyless(0); return '\n'; }
<path>.      { lcmaps_warning(PDL_ERROR, "unrecognized tokens in path name."); }
<pathcom>.   ;
<pathcom>\n  { BEGIN INITIAL; yyless(0); return '\n'; }
{nl}         { ++lineno;          }
{tokens}     { return yytext[0];  }
{trans}      { return TRANS;      }
{comment}    { BEGIN linecom;     }
<linecom>.   ;
<linecom>\n  { ++lineno; BEGIN INITIAL; }
{pdlstr}     { if ((yylval.record = (record_t*)malloc(sizeof(record_t)))) {
                 if (yytext[yyleng-2] == '\\') {
                   if (((yylval.record)->string = strdup(yytext+1))) {
                     (yylval.record)->string[yyleng-2] = '\0';
                     (yylval.record)->string[yyleng-3] = '\"';
                     (yylval.record)->lineno = lineno;
                     yyless(yyleng-1);  /*  make the next string start with "  */
                     BEGIN mq;
                     return STRING;
                   }
                 } else {
                   if (yytext[yyleng-1] == '\"') {
                     if (((yylval.record)->string = strdup(yytext+1))) {
                       (yylval.record)->string[yyleng-2] = '\0';
                       (yylval.record)->lineno = lineno;
                       return STRING;
                     }
                   }
                 }
                 lcmaps_warning(PDL_ERROR, "out of memory.");
                 return 0;
               } else
                 lcmaps_warning(PDL_ERROR, "string contains a newline.");
             }
<mq>{pdlstr} { if ((yylval.record = (record_t*)malloc(sizeof(record_t)))) {
                 if (yytext[yyleng-2] == '\\') {
                   if (((yylval.record)->string = strdup(yytext+1))) {
                     (yylval.record)->string[yyleng-2] = '\0';
                     (yylval.record)->string[yyleng-3] = '\"';
                     (yylval.record)->lineno = lineno;
                     yyless(yyleng-1);  /*  make the next string start with "  */
                     BEGIN INITIAL;
                     return STRING;
                   }
                 } else {
                   lcmaps_warning(PDL_ERROR, "Unballanced quotes.");
                   return 0;
                 }
               }
               lcmaps_warning(PDL_ERROR, "out of memory.");
               return 0;
             }
.            { lcmaps_warning(PDL_WARNING, "unrecoginzed token '%c'.", yytext[0]);
               return yytext[0];
             }
%%

#ifdef HAVE_FLEX

void delete_lex_buffer(void)
{
  yy_delete_buffer(YY_CURRENT_BUFFER);
}

#endif
