/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */


/*!
    \file   lcmaps_verify_account_from_pem.c
    \brief  This interface returns the uid, gids and poolindex, i.e. leaseid,
            using LCMAPS and takes the user credential in PEM format as input
    \author Martijn Steenbakkers for the EU DataGrid.

    This interface returns the uid, gids and poolindex, i.e. leaseid, using LCMAPS.
    As input it requires the user credential in PEM format
    -# lcmaps_verify_account_from_pem: interface function
*/

#ifndef LCMAPS_RETURN_ACCOUNT_FROM_PEM_C
#define LCMAPS_RETURN_ACCOUNT_FROM_PEM_C

/******************************************************************************
                             Include header files
******************************************************************************/
#include "lcmaps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* LCMAPS includes */
#include "lcmaps.h"
#include "lcmaps_config.h"
#include "lcmaps_log.h"
#include "lcmaps_verify_account_from_pem.h"

/******************************************************************************
                             Module specific defines
******************************************************************************/
/* Default settings */
#define LCMAPS_MAX_POLICIES 10
#define LCMAPS_POLICY_STRING "acquisition_policy"
#define LCMAPS_LOG_FILE "/var/log/lcmaps-suexec.log"
#define LCMAPS_LOG_STRING "verify_account_from_pem"
#define LCMAPS_BUFSIZ 8192
#define LCMAPS_LOG_BUFSIZ 2048
#define LCMAPS_DEFAULT_DB_FILE "lcmaps.db"
#define LCMAPS_REQUIRED_DB_PATH LCMAPS_ETC_HOME

/* Flags to indicate which type of logging is required */
#define DO_USRLOG           ((unsigned short)0x0001)
#define DO_SYSLOG           ((unsigned short)0x0002)

/******************************************************************************
                       Define module specific variables
******************************************************************************/
/* static int              do_syslog = 1; */                /* use syslog for LCMAPS logging */
/*static int              lcmaps_log_stream_exists = 0; */  /* 1 if log_stream already exists at startup */
static FILE *           lcmaps_log_stream = NULL;      /* stream for LCMAPS logging */
static char *           lcmaps_log_string = NULL;      /* (constant) log string to be prepended to log information */
static char *           lcmaps_log_file = NULL;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/
/*
static int           start_logging();
static int           log_line(int , char * , ...);
static int           log_a_string(int , char * , char *);
static int           end_logging();
*/
extern int lcmaps_tokenize( const char * command, char ** args, int * n, char * sep);
extern int lcmaps_verify_account_from_pem_va (int narg, ...);


#if 0
/******************************************************************************
Function:       start_logging
Description:    Try to open logfile, if needed.
Parameters:
Returns:        0 succes
                1 failure
******************************************************************************/
static int start_logging()
{
    if (!lcmaps_log_stream)
    {
        if ( (lcmaps_log_file) && ( (lcmaps_log_stream = fopen( lcmaps_log_file, "a")) == NULL) )
        {
            fprintf(stderr, "Cannot open logfile %s: %s\n",
                            lcmaps_log_file, strerror(errno));
            if (do_syslog)
            {
                syslog(LOG_ERR, "Cannot open logfile %s: %m", lcmaps_log_file);
            }
            return 1;
        }
    }
    else
    {
        lcmaps_log_stream_exists = 1;
    }

    return 0;
}

/******************************************************************************
Function:       log_line
Description:    Log information to file and or syslog
Parameters:
                prty:    syslog priority (if 0 don't syslog)
                fmt:     string format
Returns:        0 succes
                1 failure
******************************************************************************/
static int
log_line(int prty, char * fmt, ...)
{
    va_list pvar;
    char    buf[LCMAPS_LOG_BUFSIZ];
    int     res;

    va_start(pvar, fmt);
    res = vsnprintf(buf, LCMAPS_LOG_BUFSIZ, fmt, pvar);
    va_end(pvar);
    if ( (res >= LCMAPS_LOG_BUFSIZ) || (res < 0) )
    {
        if (do_syslog)
        {
            syslog(LOG_WARNING, "log_line() warning: log string too long (> %d)\n",
            LCMAPS_LOG_BUFSIZ);
        }
        fprintf(stderr, "log_line() warning: log string too long (> %d)\n",
                LCMAPS_LOG_BUFSIZ);
    }
    if (lcmaps_log_file)
    {
        if (lcmaps_log_stream == NULL)
        {
            if (do_syslog)
            {
                syslog(LOG_ERR, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            }
            fprintf(stderr, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            return 1;
        }
        if (lcmaps_log_string == NULL)
        {
            fprintf(lcmaps_log_stream,"LEVEL %d: %s", prty, buf);
        }
        else
        {
            fprintf(lcmaps_log_stream,"%s %d: %s", lcmaps_log_string, prty, buf);
        }
        fflush(lcmaps_log_stream);
    }
    if (do_syslog && prty)
    {
        syslog(prty, buf);
    }

    return 0;
}

/******************************************************************************
Function:       log_a_string
Description:    Log information to file and or syslog
Parameters:
                prty:       syslog priority (if 0 don't syslog)
                fmt:        string format
                the_string: the string
Returns:        0 succes
                1 failure
******************************************************************************/
static int
log_a_string(int prty, char * fmt, char * the_string)
{
    char    buf[LCMAPS_LOG_BUFSIZ];
    int     res;

    res = snprintf(buf, LCMAPS_LOG_BUFSIZ, fmt, the_string);
    if ( (res >= LCMAPS_LOG_BUFSIZ) || (res < 0) )
    {
        if (do_syslog)
        {
            syslog(LOG_WARNING, "log_line() warning: log string too long (> %d)\n",
            LCMAPS_LOG_BUFSIZ);
        }
        fprintf(stderr, "log_line() warning: log string too long (> %d)\n",
                LCMAPS_LOG_BUFSIZ);
    }
    if (lcmaps_log_file)
    {
        if (lcmaps_log_stream == NULL)
        {
            if (do_syslog)
            {
                syslog(LOG_ERR, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            }
            fprintf(stderr, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            return 1;
        }
        if (lcmaps_log_string == NULL)
        {
            fprintf(lcmaps_log_stream,"LEVEL %d: %s", prty, buf);
        }
        else
        {
            fprintf(lcmaps_log_stream,"%s %d: %s", lcmaps_log_string, prty, buf);
        }
        fflush(lcmaps_log_stream);
    }
    if (do_syslog && prty)
    {
        syslog(prty, buf);
    }

    return 0;
}

/******************************************************************************
Function:       end_logging
Description:    Close logfile
Parameters:
Returns:        0 succes
                1 failure
******************************************************************************/
static int end_logging()
{
    int retval = 0;

    if ((lcmaps_log_stream) && (lcmaps_log_stream_exists == 0))
    {
        if ( (retval = fclose(lcmaps_log_stream)) )
        {
            return retval;
        }
        else
        {
            lcmaps_log_stream = NULL;
            return 0;
        }
    }
    else
    {
        return 0;
    }
}
#endif


/******************************************************************************
Function:   lcmaps_verify_account_from_pem_va
Description:
    Verify the account mapping from the supplied user credential.

Parameters:
    narg:  the number of arguments that follow
    ap:    vararg list, that consists of
        - input:  The PEM-encoded string containing the user proxy
        - input:  a structure that contains account information:
                  uid, gids (primary and secondary) and the poolindex
Returns:
    0: success
    1: failure
******************************************************************************/
int lcmaps_verify_account_from_pem_va(
    int        narg,
    ...
)
{
    va_list                             ap;
    char *                              pem_string = NULL;        /* input variable */
    lcmaps_account_info_t               lcmaps_account;           /* input variable */
    char *                              lcmaps_policy_string = NULL;
    char *                              lcmaps_policies[LCMAPS_MAX_POLICIES];
    int                                 lcmaps_npols = LCMAPS_MAX_POLICIES;
    /* char *                              lcmaps_db_file_env = NULL; */
    /* char *                              lcmaps_db_file = NULL; */
    int                                 rc = -1;
    int                                 jj = 0;

    time_t                              mclock;
    struct tm  *                        timestamp = NULL;
    size_t                              string_size;
    char *                              lcmaps_log_string_default = NULL;
    /* struct stat                         prg_info; */   /* program info holder       */
    /* struct stat                         dir_info; */   /* directory info holder     */

    /*
     * Take log file from environment variables.
     * If not defined take the default ones
     */
    lcmaps_log_file = getenv("LCMAPS_LOG_FILE");
    lcmaps_log_file = (lcmaps_log_file ? lcmaps_log_file : LCMAPS_LOG_FILE );

    /*
     * The other environment variables should in principle not be touched,
     * because reasonably defaults are provided by the LCMAPS core.
     * An exception is made for LCMAPS_LOG_STRING, because we want to enforce
     * that it is logged which interface is used.
     * For traceability also add a timestamp.
     * setenv with overwrite=0 only overwrites if the variable is non-existant
     */
    time(&mclock);
    /* timestamp = localtime(&mclock); */
    timestamp = gmtime(&mclock);
    string_size = strlen(LCMAPS_LOG_STRING) + strlen(": ") + 19 + 1;
    lcmaps_log_string_default = (char *) malloc(string_size * sizeof(char));

    snprintf(lcmaps_log_string_default, string_size, "%s: %04d-%02d-%02d.%02d:%02d:%02d",
        LCMAPS_LOG_STRING,
        timestamp->tm_year + 1900, timestamp->tm_mon + 1, timestamp->tm_mday,
        timestamp->tm_hour, timestamp->tm_min, timestamp->tm_sec);

    setenv("LCMAPS_LOG_STRING", lcmaps_log_string_default, 0);
    lcmaps_log_string = getenv("LCMAPS_LOG_STRING");

    if (lcmaps_log_string_default != NULL)
    {
        free(lcmaps_log_string_default);
        lcmaps_log_string_default = NULL;
    }
//    /*
//     * Another exception is made for the lcmaps.db file: This interface is used by
//     * VO services (i.e. not site-controlled services), so the VO should not be able
//     * to determine which lcmaps.db file (or in the future a general lcmaps config file)
//     * they take.
//     */
//    if ( (lcmaps_db_file_env = getenv("LCMAPS_DB_FILE")) != NULL)
//    {
//        /*
//         * OK, one tries to set the location of the lcmaps.db file.
//         * We don't allow paths to be included in the filename. LCMAPS will add the
//         * site-controlled path to it.
//         * If a path is found: error
//         * If no path is found: add the default path to it and performs various checks
//         */
//        if (strchr(lcmaps_db_file_env, '/') != NULL)
//        {
//            fprintf(stderr, "%s: (LCMAPS_DB_FILE = %s)\n", lcmaps_log_string, lcmaps_db_file_env);
//            fprintf(stderr,
//                "%s: Attempt to include path to lcmaps.db file by environment variable\n", lcmaps_log_string);
//            fprintf(stderr, "%s: This is not allowed for this lcmaps interface\n", lcmaps_log_string);
//            return 1;
//        }
//    }
//    else
//    {
//        lcmaps_db_file_env = LCMAPS_DEFAULT_DB_FILE;
//    }
//    /* prepend required path */
//    lcmaps_db_file = (char *) malloc(strlen(LCMAPS_REQUIRED_DB_PATH) + 2 + strlen(lcmaps_db_file_env));
//    sprintf(lcmaps_db_file, "%s/%s", LCMAPS_REQUIRED_DB_PATH, lcmaps_db_file_env);
//
//    /*
//     * Check lcmaps.db file:
//     * - Should be in a directory owned by root, which non-writable for others.
//     * - Should be owned by root and non-writable for others.
//     * If checks are ok reset environment variable
//     */
//    /*
//     * Stat the directory of lcmaps.db and verify it is a directory, or error out.
//     */
//    if (((lstat(LCMAPS_REQUIRED_DB_PATH, &dir_info)) != 0) || !(S_ISDIR(dir_info.st_mode)))
//    {
//        fprintf(stderr, "%s: cannot stat directory: (%s)\n", lcmaps_log_string, LCMAPS_REQUIRED_DB_PATH);
//        if (lcmaps_db_file) {free(lcmaps_db_file); lcmaps_db_file = NULL;}
//        return 1;
//    }
//    /*
//     * Error out if the directory is not owned by root
//     */
//    if ((dir_info.st_uid != 0) || (dir_info.st_gid != 0))
//    {
//        fprintf(stderr, "%s: directory is not owned by root: (%s)\n", lcmaps_log_string, LCMAPS_REQUIRED_DB_PATH);
//        if (lcmaps_db_file) {free(lcmaps_db_file); lcmaps_db_file = NULL;}
//        return 1;
//    }
//    /*
//     * Error out if the directory is writable by others.
//     */
//    if (dir_info.st_mode & S_IWOTH)
//    {
//        fprintf(stderr, "%s: directory is writable by others: (%s)\n", lcmaps_log_string, LCMAPS_REQUIRED_DB_PATH);
//        if (lcmaps_db_file) {free(lcmaps_db_file); lcmaps_db_file = NULL;}
//        return 1;
//    }
//    /*
//     * Error out if we cannot stat the lcmaps.db file.
//     */
//    if (((lstat(lcmaps_db_file, &prg_info)) != 0) || (S_ISLNK(prg_info.st_mode)))
//    {
//        fprintf(stderr, "%s: cannot stat program: (%s)\n", lcmaps_log_string, lcmaps_db_file);
//        if (lcmaps_db_file) {free(lcmaps_db_file); lcmaps_db_file = NULL;}
//        return 1;
//    }
//    /*
//     * Error out if the lcmaps.db file is not owned by root
//     */
//    if ((prg_info.st_uid != 0) || (prg_info.st_gid != 0))
//    {
//        fprintf(stderr, "%s: file is not owned by root: (%s)\n", lcmaps_log_string, lcmaps_db_file);
//        if (lcmaps_db_file) {free(lcmaps_db_file); lcmaps_db_file = NULL;}
//        return 1;
//    }
//    /*
//     * Error out if the lcmaps.db file is writable by others.
//     */
//    if (prg_info.st_mode & S_IWOTH)
//    {
//        fprintf(stderr, "%s: file is writable by others: (%s)\n", lcmaps_log_string, lcmaps_db_file);
//        if (lcmaps_db_file) {free(lcmaps_db_file); lcmaps_db_file = NULL;}
//        return 1;
//    }
//    /* Everything's fine, so set (overwrite!) the environment variable. */
//    setenv("LCMAPS_DB_FILE", lcmaps_db_file, 1);
//    if (lcmaps_db_file)
//    {
//        free(lcmaps_db_file);
//        lcmaps_db_file = NULL;
//    }

    /* Parse arguments from va_list */
    va_start(ap, narg);
    if (narg == 2)
    {
        pem_string = va_arg(ap, char *);
        lcmaps_account = va_arg(ap, lcmaps_account_info_t);
    }
    else
    {
        fprintf(stderr, "%s: The number of arguments (%d) should be 2\n", lcmaps_log_string, narg);
        return 1;
    }
    va_end(ap);

#if 0
    if (start_logging())
    {
        fprintf(stderr, "%s: Cannot start logging\n", lcmaps_log_string);
        return 1;
    }
#endif

    /* Parse policy environment variable */
    for (jj=0; jj < LCMAPS_MAX_POLICIES; jj++) lcmaps_policies[jj] = 0;
    lcmaps_policy_string = getenv("LCMAPS_POLICY_NAME");
    lcmaps_policy_string = (lcmaps_policy_string ? lcmaps_policy_string : LCMAPS_POLICY_STRING );
    if ((rc = lcmaps_tokenize(lcmaps_policy_string, lcmaps_policies, &lcmaps_npols, ":")) != 0)
    {
        fprintf(stderr, "%s: Cannot parse LCMAPS_POLICY_NAME environment variable, because\n", lcmaps_log_string);
        switch (rc)
        {
            case -1:
                fprintf(stderr, "%s: of a malloc error\n", lcmaps_log_string);
                break;
            case -2:
                fprintf(stderr, "%s: the policy list is too large (max = %d)\n",
                    lcmaps_log_string, LCMAPS_MAX_POLICIES);
                break;
            case -3:
                fprintf(stderr, "%s: of a non-matching quote\n", lcmaps_log_string);
                break;
            default:
                fprintf(stderr, "%s: of an unknown error\n", lcmaps_log_string);
                break;
        }
        goto verify_account_from_pem_error;
    }

    /* First initialize LCMAPS */
    if (lcmaps_init_and_log(lcmaps_log_stream, DO_USRLOG|DO_SYSLOG))
    {
        fprintf(stderr, "%s: LCMAPS initialization failure\n", lcmaps_log_string);
        goto verify_account_from_pem_error;
    }


    /*
     * Now that we have the credential let's run (and terminate) LCMAPS !
     */
    rc = lcmaps_run_and_verify_account_from_pem(
        NULL,
        pem_string,
        lcmaps_account.uid,
        lcmaps_account.pgid_list,
        lcmaps_account.npgid,
        lcmaps_account.sgid_list,
        lcmaps_account.nsgid,
        lcmaps_account.poolindex,
        NULL,
        lcmaps_npols,
        lcmaps_policies
    );
    if (rc != 0)
    {
        lcmaps_log (1, "LCMAPS failed to verify the account mapping\n");
        if (lcmaps_term())
        {
            lcmaps_log (1, "LCMAPS termination failure\n");
            goto verify_account_from_pem_error;
        }
        goto verify_account_from_pem_error;
    }

    rc = lcmaps_term();
    if (rc)
    {
        lcmaps_log (1, "LCMAPS termination failure\n");
        goto verify_account_from_pem_error;
    }

#if 0
    if (end_logging())
    {
        fprintf(stderr, "%s: Cannot cleanly end lcmaps logging\n", lcmaps_log_string);
        goto verify_account_from_pem_error;
    }
#endif

    /* Clean policies */
    for (jj=0; jj < LCMAPS_MAX_POLICIES; jj++)
    {
        if ((lcmaps_policies[jj]) != NULL)
        {
            free(lcmaps_policies[jj]);
            lcmaps_policies[jj] = NULL;
        }
    }

    return 0;

 verify_account_from_pem_error:
    /* Clean policies */
    for (jj=0; jj < LCMAPS_MAX_POLICIES; jj++)
    {
        if ((lcmaps_policies[jj]) != NULL)
        {
            free(lcmaps_policies[jj]);
            lcmaps_policies[jj] = NULL;
        }
    }
#if 0
    if (end_logging())
    {
        fprintf(stderr, "%s: Cannot cleanly end lcmaps logging\n", lcmaps_log_string);
    }
#endif

    return 1;
}

/******************************************************************************
Function:   lcmaps_verify_account_from_pem
Description:
    Verify the account mapping from the supplied user credential.

Parameters:
    pem_string:    :  The PEM-encoded string containing the user proxy
    lcmaps_account :  A structure that contains account information: (input)
                      uid, gids (primary and secondary).
                      Please use lcmaps_account_info_clean() to clean this
                      structure after use.
Returns:
    0: success
    1: failure
******************************************************************************/
int lcmaps_verify_account_from_pem(
        char * pem_string,
        lcmaps_account_info_t lcmaps_account
)
{
    return (lcmaps_verify_account_from_pem_va(2, pem_string, lcmaps_account));
}

#endif /* LCMAPS_RETURN_ACCOUNT_FROM_PEM_C */

/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/lcmaps/src/lcmaps_verify_account_from_pem.c,v $
    $Date: 2011-03-04 21:55:13 +0100 (Fri, 04 Mar 2011) $
    $Revision: 14868 $
    $Author: dennisvd $
******************************************************************************/
